#
#

import sys
from pathlib import Path

from phlop.app import perf as p
from phlop.logger import getLogger
from phlop.testing import parallel_processor as pp
from phlop.testing import test_cases as tc

logger = getLogger(__name__)
logpath = Path(".phlop") / "perf"

"""
if you get an error about js-d3-flamegraph: https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=996839
mkdir /usr/share/d3-flame-graph
wget -O /usr/share/d3-flame-graph/d3-flamegraph-base.html https://cdn.jsdelivr.net/npm/d3-flame-graph@4/dist/templates/d3-flamegraph-base.html
"""


def perf_stat_cmd(cli_args, path, line):
    file = Path(line.split(" ")[-1]).stem
    outpath = logpath / path.stem
    outpath.mkdir(parents=True, exist_ok=True)
    return p.stat_cmd(line, p.stat_events, outpath / f"{file}.json")


def get_from_files(cli_args):
    test_batches = {}
    file_paths = list(Path(cli_args.dir).glob(cli_args.infiles))

    if not file_paths:
        raise ValueError("No load files found")

    for path in file_paths:
        with open(path, "r") as file:
            while line := file.readline():
                test_case = tc.determine_cores_for_test_case(
                    tc.TestCase(cmd=perf_stat_cmd(cli_args, path, line.rstrip()))
                )
                if test_case.cores not in test_batches:
                    test_batches[test_case.cores] = []
                test_batches[test_case.cores].append(test_case)

    return [tc.TestBatch(v, k) for k, v in test_batches.items()]


def get_remaining(cli_args):
    test_batches = {}
    path = Path(cli_args.remaining[-1]).parent
    test_case = tc.determine_cores_for_test_case(
        tc.TestCase(cmd=perf_stat_cmd(cli_args, path, " ".join(cli_args.remaining)))
    )
    test_batches[test_case.cores] = [test_case]
    return [tc.TestBatch(v, k) for k, v in test_batches.items()]


def main():
    logpath.mkdir(parents=True, exist_ok=True)

    parser = p.cli_args_parser()
    cli_args = p.verify_cli_args(parser.parse_args())
    try:
        test_batches = []
        if cli_args.infiles:
            test_batches = get_from_files(cli_args)
        else:
            test_batches = get_remaining(cli_args)

        pp.process(
            test_batches,
            n_cores=cli_args.cores,
            print_only=cli_args.print_only,
            logging=cli_args.logging,
        )

    except (Exception, SystemExit) as e:
        logger.exception(e)
        parser.print_help()
        sys.exit(1)


if __name__ == "__main__":
    main()



#
#
#
#
#


import logging
import os
import signal
import sys
import time
from dataclasses import dataclass, field
from datetime import datetime
from multiprocessing import Process, Queue

import numpy as np
import psutil
import yaml
from phlop.dict import ValDict
from phlop.proc import run_raw

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

_default_interval = 2


@dataclass
class ProcessCaptureInfo:
    cpu_load: list = field(default_factory=lambda: [])
    fds: list = field(default_factory=lambda: [])
    mem_usage: list = field(default_factory=lambda: [])
    timestamps: list = field(default_factory=lambda: [])


def cli_args_parser(description=""):
    import argparse

    _help = ValDict(
        quiet="Redirect output to /dev/null",
        interval="Seconds between process stat capture",
        yaml="write yaml file during execution",
        summary="prints summary on end",
    )
    parser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("remaining", nargs=argparse.REMAINDER)
    parser.add_argument(
        "-q", "--quiet", action="store_true", default=False, help=_help.quiet
    )
    parser.add_argument(
        "-i", "--interval", default=_default_interval, help=_help.interval
    )
    parser.add_argument("-y", "--yaml", default=None, help=_help.interval)
    parser.add_argument("-s", "--summary", action="store_true", default=True)
    return parser


def verify_cli_args(cli_args):
    try:
        cli_args.interval = int(cli_args.interval)
    except ValueError:
        raise ValueError("Interval must be an integer")
    if cli_args.yaml:
        cli_args.summary = False
    sys.argv = [sys.argv[0]]  # drop everything!
    return cli_args


def check_pid(pid):
    """Check For the existence of a unix pid."""
    if not pid:
        return False
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    return True


def signal_handler(sig, frame):
    sys.exit(0)


def bytes_as_mb(n_bytes):
    return int(n_bytes / 1024**2)


def timestamp_now():
    return datetime.utcnow().isoformat()


def now(pid):
    proc = psutil.Process(pid=pid)
    open_file = len(proc.open_files())
    mem_used_mb = bytes_as_mb(proc.memory_info().rss)
    cpu_usage = proc.cpu_percent(interval=0.1)
    return dict(open_file=open_file, mem_used_mb=mem_used_mb, cpu_usage=cpu_usage)


def capture_now(pid, data):
    now = datetime.utcnow().timestamp()
    proc = psutil.Process(pid=pid)
    data.fds += [len(proc.open_files())]
    data.mem_usage += [bytes_as_mb(proc.memory_info().rss)]
    data.cpu_load += [proc.cpu_percent(interval=0.1)]
    data.timestamps += [now]


def append_yaml(file, pid):
    keys = list(now(pid).keys())
    stats = now(pid)
    vals = {"v": ",".join([str(stats[key]) for key in keys])}
    sdump = "- " + yaml.dump(vals, indent=2)
    try:
        with open(file, "a") as f:
            f.write(sdump)
    except IOError as e:
        logger.error(f"phlop detected IO error: {e}")


def init_yaml(cli_args, pid, info):
    file = cli_args.yaml
    headers = {"headers": [str(i) for i in list(now(pid).keys())]}
    cli = {"cli_args": dict(interval=cli_args.interval)}

    try:
        with open(file, "w") as f:
            f.write("---\n")
            yaml.dump(info, f, default_flow_style=False)
            yaml.dump(cli, f, default_flow_style=False)
            yaml.dump(headers, f, default_flow_style=False)
            f.write(f"start: {timestamp_now()} \n")
            f.write("snapshots:\n")
    except IOError as e:
        logger.error(f"phlop detected IO error: {e}")


def end_yaml(file):
    try:
        with open(file, "a") as f:
            f.write(f"end: {timestamp_now()} \n")
    except IOError as e:
        logger.error(f"phlop detected IO error: {e}")


class RuntimeStatsManager:
    def __init__(self, cli_args, info={}):
        self.proc = run_raw(cli_args.remaining, quiet=cli_args.quiet)
        self.pid = self.proc.pid
        self.cli_args = cli_args

        if self.cli_args.yaml:
            init_yaml(self.cli_args, self.pid, info)

        self.pqueue = Queue()
        self.data = {}
        self.p = Process(target=RuntimeStatsManager._run, args=(self,))
        self.p.start()
        self.data = None
        if cli_args.summary:
            self.data = self.pqueue.get()

    def __del__(self):
        if check_pid(self.pid):
            os.kill(self.pid, signal.SIGINT)
        self.join()

    def join(self):
        if not self.pid:
            return
        while self.p.exitcode is None and check_pid(self.p.pid):
            time.sleep(1)
        self.pid = 0
        return self

    @staticmethod
    def _run(this):
        if this.cli_args.summary:
            data = ProcessCaptureInfo()
            signal.signal(signal.SIGINT, signal_handler)
            while check_pid(this.pid):
                try:
                    capture_now(this.pid, data)
                except psutil.AccessDenied:
                    break
                time.sleep(this.cli_args.interval)
            this.pqueue.put(data)
        else:
            while check_pid(this.pid):
                try:
                    append_yaml(this.cli_args.yaml, this.pid)
                    time.sleep(this.cli_args.interval)
                except psutil.AccessDenied:
                    break
            end_yaml(this.cli_args.yaml)


class AttachableRuntimeStatsManager:
    def __init__(self, cli_args, info={}):
        self.pid = os.getpid()
        self.cli_args = cli_args
        init_yaml(self.cli_args, self.pid, info)
        self.data = {}
        self.p = Process(target=AttachableRuntimeStatsManager._run, args=(self,))

    def start(self):
        self.daemon = True
        self.p.start()
        return self

    def kill(self):
        if self and self.p and check_pid(self.p.pid):
            os.kill(self.p.pid, signal.SIGINT)
        return self

    def join(self):
        if not self.p.pid:
            return
        while self.p.exitcode is None and check_pid(self.p.pid):
            time.sleep(1)
        self.pid = 0
        return self

    @staticmethod
    def _run(this):
        while check_pid(this.pid):
            try:
                append_yaml(this.cli_args.yaml, this.pid)
                time.sleep(this.cli_args.interval)
            except (KeyboardInterrupt, psutil.AccessDenied):
                break
        end_yaml(this.cli_args.yaml)


def print_summary(statsman):
    print("summary:")
    print("\tCPU avg:", np.average(statsman.data.cpu_load))
    print("\tCPU max:", np.max(statsman.data.cpu_load))

    print("\tFDS avg:", np.average(statsman.data.fds))
    print("\tFDS max:", np.max(statsman.data.fds))

    print("\tMEM avg:", np.average(statsman.data.mem_usage))
    print("\tMEM max:", np.max(statsman.data.mem_usage))



# valgrind.py
#

from phlop.proc import run


def run_massif(cli_args):
    outfile = cli_args.outfile if cli_args.outfile else "massif.phlop"
    cmd = f"valgrind --tool=massif --massif-out-file={outfile} " + " ".join(
        cli_args.remaining
    )
    run(cmd, capture_output=True)


def run_memcheck(cli_args):
    cmd = "valgrind " + " ".join(cli_args.remaining)
    run(cmd, capture_output=True)


def run_valgrind(cli_args):
    tool = cli_args.tool
    if tool == "massif":
        run_massif(cli_args)
    elif tool == "memcheck" or not tool:
        run_memcheck(cli_args)
    elif tool:
        raise RuntimeError(f"Phlop ERROR: Unhandled valgrind tool: {tool}")



#
#


import logging
import os

LOG_LEVEL = os.environ.get("PHLOP_LOG_LEVEL", "INFO")
log_levels = {"INFO": 20, "WARNING": 30, "ERROR": 40, "DEBUG": 10}

FORMAT = "[log: %(filename)s:%(lineno)s - %(funcName)30s() ] %(message)s"
logging.basicConfig(format=FORMAT)
# logger.setLevel(log_levels[LOG_LEVEL])


def getLogger(name, level=LOG_LEVEL):
    logger = logging.getLogger(__name__)
    level = log_levels[level] if isinstance(level, str) else level
    logger.setLevel(level)
    return logger



#
#
#

import os

from phlop.dict import ValDict
from phlop.proc import run, run_mp

# can be modified
perf_events = [
    "duration_time",
    "cycles",
    "instructions",
    "cache-references",
    "cache-misses",
    "L1-dcache-loads",
    "L1-dcache-load-misses",
]
# "perf stat" can support more events than "perf record"
stat_events = perf_events + ["bus-cycles"]


def version():
    # validated on perf version: 5.19
    proc = run("perf -v", shell=True, capture_output=True).out()
    if " " not in proc or "." not in proc:
        raise ValueError("Unparsable result from 'perf -v'")
    return [int(digit) for digit in proc.split(" ")[-1].split(".")]


def check(force_kernel_space=False):
    """perf can require some system config / read the error if thrown"""
    kernel_space_opt = "a" if force_kernel_space else ""
    run(
        f"perf stat -{kernel_space_opt}d sleep 1",
        shell=True,
        capture_output=True,
        check=True,
    )
    record("ls", [], "/tmp/perf_record_check.dat")


def parse_key(key, force_kernel_space):
    user_space_postfix = ":u"
    if key.endswith(user_space_postfix):
        if force_kernel_space:
            raise RuntimeError(f"Userspace event found {key}")
        return key[: -len(user_space_postfix)]
    return key


def parse_stat_csv(file, force_kernel_space=False):
    import csv

    comments_lines = 2  # validate
    row_val_idx, row_id_idx = 0, 2
    with open(file, newline="") as csvfile:
        [next(csvfile) for i in range(comments_lines)]  # skip headers
        return {
            parse_key(row[row_id_idx], force_kernel_space): row[row_val_idx]
            for row in csv.reader(csvfile, delimiter=",")
        }


def parse_stat_json(file):
    import json

    with open(file, newline="") as f:
        return json.loads(f)


# https://perf.wiki.kernel.org/index.php/Tutorial
# http://www.brendangregg.com/perf.html
# or run "perf list"
def events_str(events):
    if len(events) == 0:
        return ""
    return f"-e {events if isinstance(events, str) else ','.join(events)}"


def out_str(output_file):
    return "" if output_file is None else f"-o {os.path.relpath(output_file)}"


def stat_cmd(exe, events, output_file, options=""):
    return f"perf stat -j {options} {out_str(output_file)} {events_str(events)} {exe}"
    return f"perf stat -x , {options} {out_str(output_file)} {events_str(events)} {exe}"


def stat(exe, events=stat_events, output_file=None):
    return run(stat_cmd(exe, events, output_file), check=True)


def record_cmd(exe, events, output_file, options=""):
    return f"perf record {options} {out_str(output_file)} {events_str(events)} {exe}"


def record(exe, events, output_file=None):
    return run(record_cmd(exe, events, output_file), check=True)


def stat_mp(exe, events, output_files):
    return run_mp([stat_cmd(exe, events, out) for out in output_files])


def cli_args_parser(description="Perf tool"):
    import argparse

    _help = ValDict(
        dir="working directory",
        quiet="Redirect output to /dev/null",
        cores="Parallism core/thread count",
        infiles="infiles",
        print_only="Print only, no execution",
        regex="Filter out non-matching execution strings",
        logging="0=off, 1=on non zero exit code, 2=always",
        outfile="path for saved file if active",
        tool="stat/record/etc",
    )

    parser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("remaining", nargs=argparse.REMAINDER)
    parser.add_argument("-d", "--dir", default=".", help=_help.dir)
    parser.add_argument("-c", "--cores", type=int, default=1, help=_help.cores)
    parser.add_argument(
        "-p", "--print_only", action="store_true", default=False, help=_help.print_only
    )
    parser.add_argument("-i", "--infiles", default=None, help=_help.infiles)
    parser.add_argument("-r", "--regex", default=None, help=_help.regex)
    parser.add_argument("-o", "--outfile", default=None, help=_help.outfile)
    parser.add_argument("-t", "--tool", default="stat", help=_help.tool)
    parser.add_argument("--logging", type=int, default=1, help=_help.logging)
    return parser


def verify_cli_args(cli_args):
    return cli_args



#
#

import sys

from phlop.app import stats_man as sman
from phlop.logger import getLogger

logger = getLogger(__name__)

USAGE = """Stats Manager - see CPU/RAM/FD usage"""


def main():
    parser = sman.cli_args_parser(USAGE)
    cli_args = sman.verify_cli_args(parser.parse_args())
    try:
        info = dict(exe=cli_args.remaining)
        statsman = sman.RuntimeStatsManager(cli_args, info).join()
        if cli_args.summary:
            sman.print_summary(statsman)
    except (Exception, SystemExit) as e:
        logger.exception(e)
        parser.print_help()
        sys.exit(1)


if __name__ == "__main__":
    main()



#
#
#
#


import os
import subprocess
import time

from phlop.os import pushd, write_to_file
from phlop.string import decode_bytes


class RunTimer:
    def __init__(
        self,
        cmd,
        shell=False,
        capture_output=True,
        check=False,
        print_cmd=True,
        env: dict = {},  # dict[str, str] # eventually
        working_dir=None,
        log_file_path=None,
        logging=2,
        popen=True,
        **kwargs,
    ):
        self.cmd = cmd
        self.stdout = ""
        self.stderr = ""
        self.run_time = None
        self.logging = logging
        self.log_file_path = log_file_path
        self.capture_output = capture_output
        benv = os.environ.copy()
        benv.update(env)
        ekwargs = dict(
            shell=shell,
            env=benv,
            close_fds=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        if not capture_output and log_file_path:
            ekwargs.update(
                dict(
                    stdout=open(f"{log_file_path}.stdout", "w"),
                    stderr=open(f"{log_file_path}.stderr", "w"),
                ),
            )
        else:
            ekwargs.update(
                dict(stdout=subprocess.PIPE, stderr=subprocess.PIPE),
            )

        def go():
            if popen:
                self._popen(**ekwargs, **kwargs)
            else:
                self._run(
                    check=check, capture_output=capture_output, **ekwargs, **kwargs
                )

        if working_dir:
            with pushd(working_dir):
                go()
        else:
            go()

    def _locals(self):
        return self.capture_output, self.log_file_path, self.logging

    def _run(self, **kwargs):
        capture_output, log_file_path, logging = self._locals()
        try:
            start = time.time()
            self.run = subprocess.run(self.cmd, **kwargs)
            self.run_time = time.time() - start
            self.exitcode = self.run.returncode
            if logging == 2 and capture_output:
                self.stdout = decode_bytes(self.run.stdout)
                self.stderr = decode_bytes(self.run.stderr)
        except subprocess.CalledProcessError as e:
            # only triggers on failure if check=True
            self.run_time = time.time() - start
            self.exitcode = e.returncode
            if logging >= 1 and capture_output:
                self.stdout = decode_bytes(e.stdout)
                self.stderr = decode_bytes(e.stderr)
                logging = 2  # force logging as exception occurred
        if logging == 2 and capture_output and log_file_path:
            write_to_file(f"{log_file_path}.stdout", self.stdout)
            write_to_file(f"{log_file_path}.stderr", self.stderr)

    def _popen(self, **kwargs):
        capture_output, log_file_path, logging = self._locals()
        start = time.time()
        p = subprocess.Popen(self.cmd, **kwargs)
        self.stdout, self.stderr = p.communicate()
        self.run_time = time.time() - start
        self.exitcode = p.returncode
        if not capture_output and log_file_path:
            kwargs["stdout"].close()
            kwargs["stderr"].close()
        elif capture_output:
            p.stdout.close()
            p.stderr.close()
        p = None

        if self.exitcode > 0 and capture_output:
            logging = 2  # force logging as exception occurred
        if logging == 2 and capture_output:
            self.stdout = decode_bytes(self.stdout)
            self.stderr = decode_bytes(self.stderr)
        if logging == 2 and capture_output and log_file_path:
            write_to_file(f"{log_file_path}.stdout", self.stdout)
            write_to_file(f"{log_file_path}.stderr", self.stderr)

    def out(self, ignore_exit_code=False):
        if not ignore_exit_code and self.exitcode > 0:
            raise RuntimeError(f"phlop.RunTimer error: {self.stderr}")
        return self.stdout



#
#

import sys

from phlop.app import valgrind as vg
from phlop.dict import ValDict
from phlop.logger import getLogger

logger = getLogger(__name__)


def cli_args_parser(description="valgrinding"):
    import argparse

    _help = ValDict(
        quiet="Redirect output to /dev/null",
        tool="which valgrind tool to pick (massif/etc)",
        outfile="outfile path base if available",
    )
    parser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("remaining", nargs=argparse.REMAINDER)
    parser.add_argument(
        "-q", "--quiet", action="store_true", default=False, help=_help.quiet
    )
    parser.add_argument("-t", "--tool", default="", help=_help.tool)
    parser.add_argument("-o", "--outfile", default="", help=_help.outfile)
    return parser


def verify_cli_args(cli_args):
    return cli_args


def main():
    parser = cli_args_parser()
    cli_args = verify_cli_args(parser.parse_args())
    try:
        vg.run_valgrind(cli_args)

    except (Exception, SystemExit) as e:
        logger.exception(e)
        parser.print_help()
        sys.exit(1)


if __name__ == "__main__":
    main()
